use std::{future::Future, pin::Pin};

use futures::join;

use crate::func::*;

pub struct FutureClass;

impl WeakFunctor for FutureClass {
    type F<'a, A: 'a> = Pin<Box<dyn 'a + Future<Output = A>>>;
}

impl Functor for FutureClass {
    fn fmap<'a, A: 'a, B: 'a>(f: impl 'a + FnOnce(A) -> B, fa: Self::F<'a, A>) -> Self::F<'a, B> {
        Box::pin(async { f(fa.await) })
    }

    fn replace<'a, A: 'a, B: 'a>(fa: Self::F<'a, A>, b: B) -> Self::F<'a, B> {
        Box::pin(async {
            fa.await;
            b
        })
    }
}

impl ApplicativeSeq for FutureClass {
    fn seq<'a, A: 'a, B: 'a>(
        ff: Self::F<'a, impl 'a + FnOnce(A) -> B>,
        fa: Self::F<'a, A>,
    ) -> Self::F<'a, B> {
        Box::pin(async {
            let (f, a) = join!(ff, fa);
            f(a)
        })
    }
}

impl ApplicativeLA2 for FutureClass {
    fn la2<'a, A: 'a, B: 'a, C: 'a>(
        f: impl 'a + FnOnce(A, B) -> C,
        fa: Self::F<'a, A>,
        fb: Self::F<'a, B>,
    ) -> Self::F<'a, C> {
        Box::pin(async {
            let (a, b) = join!(fa, fb);
            f(a, b)
        })
    }
}

impl Applicative for FutureClass {
    fn pure<'a, A: 'a>(a: A) -> Self::F<'a, A> {
        Box::pin(async { a })
    }

    fn discard_first<'a, A: 'a, B: 'a>(fa: Self::F<'a, A>, fb: Self::F<'a, B>) -> Self::F<'a, B> {
        Box::pin(async { join!(fa, fb).1 })
    }

    fn discard_second<'a, A: 'a, B: 'a>(fa: Self::F<'a, A>, fb: Self::F<'a, B>) -> Self::F<'a, A> {
        Box::pin(async { join!(fa, fb).0 })
    }
}

impl Monad for FutureClass {
    fn bind<'a, A: 'a, B: 'a>(
        fa: Self::F<'a, A>,
        f: impl 'a + FnOnce(A) -> Self::F<'a, B>,
    ) -> Self::F<'a, B> {
        Box::pin(async { f(fa.await).await })
    }

    fn join<'a, A: 'a>(ffa: Self::F<'a, Self::F<'a, A>>) -> Self::F<'a, A> {
        Box::pin(async { ffa.await.await })
    }
}
